# -*- coding: utf-8 -*-
from __future__ import absolute_import, print_function, division
import itertools
from multiprocessing import Pool
from multiprocessing.pool import ThreadPool


import numpy as np
import pytest


from numcodecs import blosc
from numcodecs.blosc import Blosc
from numcodecs.tests.common import (check_encode_decode, check_config,
                                    check_backwards_compatibility,
                                    check_err_decode_object_buffer,
                                    check_err_encode_object_buffer,
                                    check_max_buffer_size)


codecs = [
    Blosc(shuffle=Blosc.SHUFFLE),
    Blosc(clevel=0, shuffle=Blosc.SHUFFLE),
    Blosc(cname='lz4', shuffle=Blosc.SHUFFLE),
    Blosc(cname='lz4', clevel=1, shuffle=Blosc.NOSHUFFLE),
    Blosc(cname='lz4', clevel=5, shuffle=Blosc.SHUFFLE),
    Blosc(cname='lz4', clevel=9, shuffle=Blosc.BITSHUFFLE),
    Blosc(cname='zlib', clevel=1, shuffle=0),
    Blosc(cname='zstd', clevel=1, shuffle=1),
    Blosc(cname='blosclz', clevel=1, shuffle=2),
    Blosc(cname='snappy', clevel=1, shuffle=2),
    Blosc(shuffle=Blosc.SHUFFLE, blocksize=0),
    Blosc(shuffle=Blosc.SHUFFLE, blocksize=2**8),
    Blosc(cname='lz4', clevel=1, shuffle=Blosc.NOSHUFFLE, blocksize=2**8),
]


# mix of dtypes: integer, float, bool, string
# mix of shapes: 1D, 2D, 3D
# mix of orders: C, F
arrays = [
    np.arange(1000, dtype='i4'),
    np.linspace(1000, 1001, 1000, dtype='f8'),
    np.random.normal(loc=1000, scale=1, size=(100, 10)),
    np.random.randint(0, 2, size=1000, dtype=bool).reshape(100, 10, order='F'),
    np.random.choice([b'a', b'bb', b'ccc'], size=1000).reshape(10, 10, 10),
    np.random.randint(0, 2**60, size=1000, dtype='u8').view('M8[ns]'),
    np.random.randint(0, 2**60, size=1000, dtype='u8').view('m8[ns]'),
    np.random.randint(0, 2**25, size=1000, dtype='u8').view('M8[m]'),
    np.random.randint(0, 2**25, size=1000, dtype='u8').view('m8[m]'),
    np.random.randint(-2**63, -2**63 + 20, size=1000, dtype='i8').view('M8[ns]'),
    np.random.randint(-2**63, -2**63 + 20, size=1000, dtype='i8').view('m8[ns]'),
    np.random.randint(-2**63, -2**63 + 20, size=1000, dtype='i8').view('M8[m]'),
    np.random.randint(-2**63, -2**63 + 20, size=1000, dtype='i8').view('m8[m]'),
]


def test_encode_decode():
    for arr, codec in itertools.product(arrays, codecs):
        check_encode_decode(arr, codec)


def test_config():
    codec = Blosc(cname='zstd', clevel=3, shuffle=1)
    check_config(codec)
    codec = Blosc(cname='lz4', clevel=1, shuffle=2, blocksize=2**8)
    check_config(codec)


def test_repr():
    expect = "Blosc(cname='zstd', clevel=3, shuffle=SHUFFLE, blocksize=0)"
    actual = repr(Blosc(cname='zstd', clevel=3, shuffle=Blosc.SHUFFLE, blocksize=0))
    assert expect == actual
    expect = "Blosc(cname='lz4', clevel=1, shuffle=NOSHUFFLE, blocksize=256)"
    actual = repr(Blosc(cname='lz4', clevel=1, shuffle=Blosc.NOSHUFFLE, blocksize=256))
    assert expect == actual
    expect = "Blosc(cname='zlib', clevel=9, shuffle=BITSHUFFLE, blocksize=512)"
    actual = repr(Blosc(cname='zlib', clevel=9, shuffle=Blosc.BITSHUFFLE, blocksize=512))
    assert expect == actual
    expect = "Blosc(cname='blosclz', clevel=5, shuffle=AUTOSHUFFLE, blocksize=1024)"
    actual = repr(Blosc(cname='blosclz', clevel=5, shuffle=Blosc.AUTOSHUFFLE,
                        blocksize=1024))
    assert expect == actual


def test_eq():
    assert Blosc() == Blosc()
    assert Blosc(cname='lz4') != Blosc(cname='zstd')
    assert Blosc(clevel=1) != Blosc(clevel=9)
    assert Blosc(cname='lz4') != 'foo'


def test_compress_blocksize():
    arr = np.arange(1000, dtype='i4')

    for use_threads in True, False, None:
        blosc.use_threads = use_threads

        # default blocksize
        enc = blosc.compress(arr, b'lz4', 1, Blosc.NOSHUFFLE)
        _, _, blocksize = blosc.cbuffer_sizes(enc)
        assert blocksize > 0

        # explicit default blocksize
        enc = blosc.compress(arr, b'lz4', 1, Blosc.NOSHUFFLE, 0)
        _, _, blocksize = blosc.cbuffer_sizes(enc)
        assert blocksize > 0

        # custom blocksize
        for bs in 2**7, 2**8:
            enc = blosc.compress(arr, b'lz4', 1, Blosc.NOSHUFFLE, bs)
            _, _, blocksize = blosc.cbuffer_sizes(enc)
            assert blocksize == bs


def test_compress_complib():
    arr = np.arange(1000, dtype='i4')
    expected_complibs = {
        'lz4': 'LZ4',
        'lz4hc': 'LZ4',
        'blosclz': 'BloscLZ',
        'snappy': 'Snappy',
        'zlib': 'Zlib',
        'zstd': 'Zstd',
    }
    for use_threads in True, False, None:
        blosc.use_threads = use_threads
        for cname in blosc.list_compressors():
            enc = blosc.compress(arr, cname.encode(), 1, Blosc.NOSHUFFLE)
            complib = blosc.cbuffer_complib(enc)
            expected_complib = expected_complibs[cname]
            assert complib == expected_complib
        with pytest.raises(ValueError):
            # capitalized cname
            blosc.compress(arr, b'LZ4', 1)
        with pytest.raises(ValueError):
            # bad cname
            blosc.compress(arr, b'foo', 1)


def test_compress_metainfo():
    for dtype in 'i1', 'i2', 'i4', 'i8':
        arr = np.arange(1000, dtype=dtype)
        for shuffle in Blosc.NOSHUFFLE, Blosc.SHUFFLE, Blosc.BITSHUFFLE:
            for use_threads in True, False, None:
                blosc.use_threads = use_threads
                for cname in blosc.list_compressors():
                    enc = blosc.compress(arr, cname.encode(), 1, shuffle)
                    typesize, did_shuffle, _ = blosc.cbuffer_metainfo(enc)
                    assert typesize == arr.dtype.itemsize
                    assert did_shuffle == shuffle


def test_compress_autoshuffle():
    arr = np.arange(8000)
    for dtype in 'i1', 'i2', 'i4', 'i8', 'f2', 'f4', 'f8', 'bool', 'S10':
        varr = arr.view(dtype)
        for use_threads in True, False, None:
            blosc.use_threads = use_threads
            for cname in blosc.list_compressors():
                enc = blosc.compress(varr, cname.encode(), 1, Blosc.AUTOSHUFFLE)
                typesize, did_shuffle, _ = blosc.cbuffer_metainfo(enc)
                assert typesize == varr.dtype.itemsize
                if typesize == 1:
                    assert did_shuffle == Blosc.BITSHUFFLE
                else:
                    assert did_shuffle == Blosc.SHUFFLE


def test_config_blocksize():
    # N.B., we want to be backwards compatible with any config where blocksize is not
    # explicitly stated

    # blocksize not stated
    config = dict(cname='lz4', clevel=1, shuffle=Blosc.SHUFFLE)
    codec = Blosc.from_config(config)
    assert codec.blocksize == 0

    # blocksize stated
    config = dict(cname='lz4', clevel=1, shuffle=Blosc.SHUFFLE, blocksize=2**8)
    codec = Blosc.from_config(config)
    assert codec.blocksize == 2**8


def test_backwards_compatibility():
    check_backwards_compatibility(Blosc.codec_id, arrays, codecs)


def _encode_worker(data):
    compressor = Blosc(cname='zlib', clevel=9, shuffle=Blosc.SHUFFLE)
    enc = compressor.encode(data)
    return enc


def _decode_worker(enc):
    compressor = Blosc()
    data = compressor.decode(enc)
    return data


def test_multiprocessing():
    data = np.arange(1000000)
    enc = _encode_worker(data)

    try:
        for use_threads in None, True, False:
            blosc.use_threads = use_threads

            # test with process pool and thread pool
            for pool in Pool(5), ThreadPool(5):

                # test encoding
                enc_results = pool.map(_encode_worker, [data] * 5)
                assert all([len(enc) == len(e) for e in enc_results])

                # test decoding
                dec_results = pool.map(_decode_worker, [enc] * 5)
                assert all([data.nbytes == len(d) for d in dec_results])

                # tidy up
                pool.close()
                pool.join()

    finally:
        blosc.use_threads = None  # restore default


def test_err_decode_object_buffer():
    check_err_decode_object_buffer(Blosc())


def test_err_encode_object_buffer():
    check_err_encode_object_buffer(Blosc())


def test_decompression_error_handling():
    for codec in codecs:
        with pytest.raises(RuntimeError):
            codec.decode(bytearray())
        with pytest.raises(RuntimeError):
            codec.decode(bytearray(0))


def test_max_buffer_size():
    for codec in codecs:
        assert codec.max_buffer_size == 2**31 - 1
        check_max_buffer_size(codec)



# -*- coding: utf-8 -*-
from __future__ import absolute_import, print_function, division
import unittest
import warnings


import numpy as np


try:
    from numcodecs.msgpacks import LegacyMsgPack, MsgPack
    default_codec = MsgPack()
    # N.B., legacy codec is broken, see tests below. Also legacy code generates
    # PendingDeprecationWarning due to use of encoding argument, which we ignore here
    # as not relevant.
    legacy_codec = LegacyMsgPack()
except ImportError:  # pragma: no cover
    raise unittest.SkipTest("msgpack not available")


from numcodecs.tests.common import (check_config, check_repr, check_encode_decode_array,
                                    check_backwards_compatibility, greetings)
from numcodecs.compat import text_type, binary_type, PY2


# object array with strings
# object array with mix strings / nans
# object array with mix of string, int, float
# ...
arrays = [
    np.array([u'foo', u'bar', u'baz'] * 300, dtype=object),
    np.array([[u'foo', u'bar', np.nan]] * 300, dtype=object),
    np.array([u'foo', 1.0, 2] * 300, dtype=object),
    np.arange(1000, dtype='i4'),
    np.array([u'foo', u'bar', u'baz'] * 300),
    np.array([u'foo', [u'bar', 1.0, 2], {u'a': u'b', u'c': 42}] * 300, dtype=object),
    np.array(greetings * 100),
    np.array(greetings * 100, dtype=object),
    np.array([b'foo', b'bar', b'baz'] * 300, dtype=object),
    np.array([g.encode('utf-8') for g in greetings] * 100, dtype=object),
]


legacy_arrays = arrays[:8]


def test_encode_decode():

    for arr in arrays:
        check_encode_decode_array(arr, default_codec)

    with warnings.catch_warnings():
        warnings.simplefilter('ignore', PendingDeprecationWarning)
        for arr in legacy_arrays:
            check_encode_decode_array(arr, legacy_codec)


def test_config():
    for codec in [default_codec, legacy_codec]:
        check_config(codec)


def test_repr():
    check_repr("MsgPack(raw=False, use_bin_type=True, use_single_float=False)")
    check_repr("MsgPack(raw=True, use_bin_type=False, use_single_float=True)")
    check_repr("LegacyMsgPack(encoding='utf-8')")
    check_repr("LegacyMsgPack(encoding='ascii')")


def test_backwards_compatibility():
    check_backwards_compatibility(default_codec.codec_id, arrays, [default_codec])
    with warnings.catch_warnings():
        warnings.simplefilter('ignore', PendingDeprecationWarning)
        check_backwards_compatibility(legacy_codec.codec_id, legacy_arrays,
                                      [legacy_codec])


def test_non_numpy_inputs():
    # numpy will infer a range of different shapes and dtypes for these inputs.
    # Make sure that round-tripping through encode preserves this.
    data = [
        [0, 1],
        [[0, 1], [2, 3]],
        [[0], [1], [2, 3]],
        [[[0, 0]], [[1, 1]], [[2, 3]]],
        [u"1"],
        [u"11", u"11"],
        [u"11", u"1", u"1"],
        [{}],
        [{u"key": u"value"}, [u"list", u"of", u"strings"]],
        [b"1"],
        [b"11", b"11"],
        [b"11", b"1", b"1"],
        [{b"key": b"value"}, [b"list", b"of", b"strings"]],
    ]
    for input_data in data:
        actual = default_codec.decode(default_codec.encode(input_data))
        expect = np.array(input_data)
        assert expect.shape == actual.shape
        assert np.array_equal(expect, actual)


def test_legacy_codec_broken():
    # Simplest demonstration of why the MsgPack codec needed to be changed.
    # The LegacyMsgPack codec didn't include shape information in the serialised
    # bytes, which gave different shapes in the input and output under certain
    # circumstances.
    a = np.empty(2, dtype=object)
    a[0] = [0, 1]
    a[1] = [2, 3]
    codec = LegacyMsgPack()
    with warnings.catch_warnings():
        warnings.simplefilter('ignore', PendingDeprecationWarning)
        b = codec.decode(codec.encode(a))
    assert a.shape == (2,)
    assert b.shape == (2, 2)
    assert not np.array_equal(a, b)

    # Now show that the MsgPack codec handles this case properly.
    codec = MsgPack()
    b = codec.decode(codec.encode(a))
    assert np.array_equal(a, b)
    assert a.shape == b.shape


def test_encode_decode_shape_dtype_preserved():
    for arr in arrays:
        actual = default_codec.decode(default_codec.encode(arr))
        assert arr.shape == actual.shape
        assert arr.dtype == actual.dtype


def test_bytes():
    # test msgpack behaviour with bytes and str (unicode)
    bytes_arr = np.array([b'foo', b'bar', b'baz'], dtype=object)
    unicode_arr = np.array([u'foo', u'bar', u'baz'], dtype=object)

    # raw=False (default)
    codec = MsgPack()
    # works for bytes array, round-trips bytes to bytes
    b = codec.decode(codec.encode(bytes_arr))
    assert np.array_equal(bytes_arr, b)
    assert isinstance(b[0], binary_type)
    assert b[0] == b'foo'
    # works for unicode array, round-trips unicode to unicode
    b = codec.decode(codec.encode(unicode_arr))
    assert np.array_equal(unicode_arr, b)
    assert isinstance(b[0], text_type)
    assert b[0] == u'foo'

    # raw=True
    codec = MsgPack(raw=True)
    # works for bytes array, round-trips bytes to bytes
    b = codec.decode(codec.encode(bytes_arr))
    assert np.array_equal(bytes_arr, b)
    assert isinstance(b[0], binary_type)
    assert b[0] == b'foo'
    # broken for unicode array, round-trips unicode to bytes
    b = codec.decode(codec.encode(unicode_arr))
    if PY2:  # pragma: py3 no cover
        # PY2 considers b'foo' and u'foo' to be equal
        assert np.array_equal(unicode_arr, b)
    else:  # pragma: py2 no cover
        assert not np.array_equal(unicode_arr, b)
    assert isinstance(b[0], binary_type)
    assert b[0] == b'foo'

    # legacy codec
    codec = LegacyMsgPack()
    with warnings.catch_warnings():
        warnings.simplefilter('ignore', PendingDeprecationWarning)
        # broken for bytes array, round-trips bytes to unicode
        b = codec.decode(codec.encode(bytes_arr))
        if PY2:  # pragma: py3 no cover
            # PY2 considers b'foo' and u'foo' to be equal
            assert np.array_equal(unicode_arr, b)
        else:  # pragma: py2 no cover
            assert not np.array_equal(bytes_arr, b)
        assert isinstance(b[0], text_type)
        assert b[0] == u'foo'
        # works for unicode array, round-trips unicode to unicode
        b = codec.decode(codec.encode(unicode_arr))
        assert np.array_equal(unicode_arr, b)
        assert isinstance(b[0], text_type)
        assert b[0] == u'foo'



print("import: 'numcodecs'")
import numcodecs

print("import: 'numcodecs.blosc'")
import numcodecs.blosc

print("import: 'numcodecs.lz4'")
import numcodecs.lz4

print("import: 'numcodecs.zstd'")
import numcodecs.zstd




# -*- coding: utf-8 -*-
from __future__ import absolute_import, print_function, division


import numpy as np


from .abc import Codec
from .compat import ensure_ndarray, ndarray_copy


class PackBits(Codec):
    """Codec to pack elements of a boolean array into bits in a uint8 array.

    Examples
    --------
    >>> import numcodecs
    >>> import numpy as np
    >>> codec = numcodecs.PackBits()
    >>> x = np.array([True, False, False, True], dtype=bool)
    >>> y = codec.encode(x)
    >>> y
    array([  4, 144], dtype=uint8)
    >>> z = codec.decode(y)
    >>> z
    array([ True, False, False,  True])

    Notes
    -----
    The first element of the encoded array stores the number of bits that
    were padded to complete the final byte.

    """

    codec_id = 'packbits'

    def __init__(self):
        pass

    def encode(self, buf):

        # normalise input
        arr = ensure_ndarray(buf).view(bool)

        # flatten to simplify implementation
        arr = arr.reshape(-1, order='A')

        # determine size of packed data
        n = arr.size
        n_bytes_packed = (n // 8)
        n_bits_leftover = n % 8
        if n_bits_leftover > 0:
            n_bytes_packed += 1

        # setup output
        enc = np.empty(n_bytes_packed + 1, dtype='u1')

        # store how many bits were padded
        if n_bits_leftover:
            n_bits_padded = 8 - n_bits_leftover
        else:
            n_bits_padded = 0
        enc[0] = n_bits_padded

        # apply encoding
        enc[1:] = np.packbits(arr)

        return enc

    def decode(self, buf, out=None):

        # normalise input
        enc = ensure_ndarray(buf).view('u1')

        # flatten to simplify implementation
        enc = enc.reshape(-1, order='A')

        # find out how many bits were padded
        n_bits_padded = int(enc[0])

        # apply decoding
        dec = np.unpackbits(enc[1:])

        # remove padded bits
        if n_bits_padded:
            dec = dec[:-n_bits_padded]

        # view as boolean array
        dec = dec.view(bool)

        # handle destination
        return ndarray_copy(dec, out)



# -*- coding: utf-8 -*-
from __future__ import absolute_import, print_function, division


import numpy as np
from numpy.testing import assert_array_equal
import pytest


from numcodecs.delta import Delta
from numcodecs.tests.common import (check_encode_decode, check_config, check_repr,
                                    check_backwards_compatibility)


# mix of dtypes: integer, float
# mix of shapes: 1D, 2D, 3D
# mix of orders: C, F
arrays = [
    np.arange(1000, dtype='i4'),
    np.linspace(1000, 1001, 1000, dtype='f4').reshape(100, 10),
    np.random.normal(loc=1000, scale=1, size=(10, 10, 10)).astype('f8'),
    np.random.randint(0, 200, size=1000, dtype='u2').reshape(100, 10, order='F'),
]


def test_encode_decode():
    for arr in arrays:
        codec = Delta(dtype=arr.dtype)
        check_encode_decode(arr, codec)


def test_encode():
    dtype = 'i8'
    astype = 'i4'
    codec = Delta(dtype=dtype, astype=astype)
    arr = np.arange(10, 20, 1, dtype=dtype)
    expect = np.array([10] + ([1] * 9), dtype=astype)
    actual = codec.encode(arr)
    assert_array_equal(expect, actual)
    assert np.dtype(astype) == actual.dtype


def test_config():
    codec = Delta(dtype='<i4', astype='<i2')
    check_config(codec)


def test_repr():
    check_repr("Delta(dtype='<i4', astype='<i2')")


def test_backwards_compatibility():
    for arr in arrays:
        codec = Delta(dtype=arr.dtype)
        check_backwards_compatibility(Delta.codec_id, [arr], [codec], prefix=str(arr.dtype))


def test_errors():
    with pytest.raises(ValueError):
        Delta(dtype=object)
    with pytest.raises(ValueError):
        Delta(dtype='i8', astype=object)



# -*- coding: utf-8 -*-
from __future__ import absolute_import, print_function, division
import zlib as _zlib


from .abc import Codec
from .compat import ndarray_copy, ensure_contiguous_ndarray


class Zlib(Codec):
    """Codec providing compression using zlib via the Python standard library.

    Parameters
    ----------
    level : int
        Compression level.

    """

    codec_id = 'zlib'

    def __init__(self, level=1):
        self.level = level

    def encode(self, buf):

        # normalise inputs
        buf = ensure_contiguous_ndarray(buf)

        # do compression
        return _zlib.compress(buf, self.level)

    # noinspection PyMethodMayBeStatic
    def decode(self, buf, out=None):

        # normalise inputs
        buf = ensure_contiguous_ndarray(buf)
        if out is not None:
            out = ensure_contiguous_ndarray(out)

        # do decompression
        dec = _zlib.decompress(buf)

        # handle destination - Python standard library zlib module does not
        # support direct decompression into buffer, so we have to copy into
        # out if given
        return ndarray_copy(dec, out)
